/***********************************************************************
 * File: data.c
 * 	(Extracted from
 *	   "$Id: data.c,v 1.1 2008/08/09 16:55:45 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C source code implements a test-program of the converter set
 * for the CSV-files "csv_converter.c".
 *
 * Copyright (C) 2004-2007 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/


/* include definitions */
#include "csv_converter.h"

/*+++
 * Test above functions...
 +++*/
static struct test_table_t {
	struct test_line_t {
		char *strings[13];
			/* pointing to allocated string area */
	} lines[16];	/* pointing to allocated test_line_t area */
} table = {
		  {
#if 0 /* { DBG***/
		  {	{ "", "", ""
				, "", "", ""
				, "", "", ""
				, "", "", ""
				, NULL /* end of lines */
			} } /* pattern 1 */
		, {	{ "a", "\"", ","
				, "a\na", "a\n\"", "a\n,"
				, "\"\na", "\"\n\"", "\"\n,"
				, ",\na", ",\n\"", ",\n,"
				, NULL /* end of lines */
			} } /* pattern 2 */
		, {	{ "aa", "a\"", "a,", "a\n"
				, "\"\"", "\"a", "\",", "\"\n"
				, ",,", ",\"", ",a", ",\n"
				, NULL /* end of lines */
			} } /* pattern 3 */
		, {	{ "\"", ",", "a\na"
				, "a\n\"", "a\n,", "\"\na"
				, "\"\n\"", "\"\n,", ",\na"
				, ",\n\"", ",\n,", "a"
				, NULL /* end of lines */
			} } /* pattern 4 */
		, {	{ ",", "a\na", "a\n\"" 
				, "a\n,", "\"\na", "\"\n\"" 
				, "\"\n,", ",\na", ",\n\"" 
				, ",\n,", "a", "\""
				, NULL /* end of lines */
			} } /* pattern 5 */
		, {	{ "a\na", "a\n\"", "a\n," 
				, "\"\na", "\"\n\"", "\"\n," 
				, ",\na", ",\n\"", ",\n,"
				, "a", "\"", ","
				, NULL /* end of lines */
			} } /* pattern 6 */
		, {	{ "a\n\"", "a\n,", "\"\na"
				, "\"\n\"", "\"\n,", ",\na"
				, ",\n\"", ",\n,", "a"
				, "\"", ",", "a\na"
				, NULL /* end of lines */
			} } /* pattern 7 */
		, {	{ "a\n,", "\"\na", "\"\n\"" 
				, "\"\n,", ",\na", ",\n\"" 
				, ",\n,", "a", "\""
				, ",", "a\na", "a\n\"" 
				, NULL /* end of lines */
			} } /* pattern 8 */
		, {	{ "\"\na", "\"\n\"", "\"\n," 
				, ",\na", ",\n\"", ",\n,"
				, "a", "\"", ","
				, "a\na", "a\n\"", "a\n," 
				, NULL /* end of lines */
			} } /* pattern 9 */
		, {	{ "\"\n\"", "\"\n,", ",\na"
				, ",\n\"", ",\n,", "a"
				, "\"", ",", "a\na"
				, "a\n\"", "a\n,", "\"\na"
				, NULL /* end of lines */
			} } /* pattern 10 */
		, {	{ "\"\n,", ",\na", ",\n\"" 
				, ",\n,", "a", "\""
				, ",", "a\na", "a\n\"" 
				, "a\n,", "\"\na", "\"\n\"" 
				, NULL /* end of lines */
			} } /* pattern 11 */
		, {	{ ",\na", ",\n\"", ",\n,"
				, "a", "\"", ","
				, "a\na", "a\n\"", "a\n," 
				, "\"\na", "\"\n\"", "\"\n," 
				, NULL /* end of lines */
			} } /* pattern 12 */
		, {	{ ",\n\"", ",\n,", "a"
				, "\"", ",", "a\na"
				, "a\n\"", "a\n,", "\"\na"
				, "\"\n\"", "\"\n,", ",\na"
				, NULL /* end of lines */
			} } /* pattern 13 */
		, {	{ ",\n,", "a", "\""
				, ",", "a\na", "a\n\"" 
				, "a\n,", "\"\na", "\"\n\"" 
				, "\"\n,", ",\na", ",\n\"" 
				, NULL /* end of lines */
			} } /* pattern 14 */
		,
#endif /* } DBG***/
		{	{ "a", "\"", ","
				, "a\na", "a\n\"", "a\n," 
				, "\"\na", "\"\n\"", "\"\n," 
				, ",\na", ",\n\"", ",\n,"
				, NULL /* end of lines */
			} } /* pattern 15 */
		, { { NULL, } } /* end of patterns */
		}
	};

static char	*csv_files[] = {
	  "what_means_CSV_p/empty_cell.csv"
	, "what_means_CSV_p/only_1_cell.csv"
	, "what_means_CSV_p/only_CR_cell.csv"
	, "what_means_CSV_p/"
		"alphabet_double-quotation_comma_and_new-line_cells.csv"
	, "what_means_CSV_p/zero_size.csv"
	, "what_means_CSV_p/"
		"broken_"
		"alphabet_double-quotation_comma_and_new-line_cells.csv"
	, "what_means_CSV_p/"
		"alphabet_double-quotation_comma_and_new-line_cells.txt"
	};
#	define CNT_csv_files (sizeof(csv_files) / sizeof(csv_files[0]))

static void show_table(struct test_table_t *t)
{
	char	*s;
	int	i, j;

	for (i = 0; t->lines[i].strings[0] != NULL; ++i) {
		printf("*** pattern %d ***\n", i + 1);
		for (j = 0; (s = t->lines[i].strings[j]) != NULL; ++j) {
			printf("\tstrings[%d]: {%s}\n", j, s);
		}
	}

	return;
}

/* test for string_to_csv_string(construct char *string) */
static void test_string_to_csv_string(struct test_table_t *t)
{
	char	*s;
	int	i, j, err_cnt = 0;
	size_t	csv_buf_sz = (size_t)0;
	char	*csv_str = NULL;

	printf("+++ test_string_to_csv_string(%p) starts...\n"
		,                                 (void *)t
		);
	for (i = 0; t->lines[i].strings[0] != NULL; ++i) {
		printf("*** pattern %d ***\n", i + 1);
		for (j = 0; (s = t->lines[i].strings[j]) != NULL; ++j) {
			printf("\tC-strings[%d]: {%s} "
				,               j,    s
				); (void)fflush(stdout);

			/* covert it! */
			csv_str = string_to_csv_string(s
						, csv_str
						, &csv_buf_sz
						);

			printf("=> CSV-String {%p (size = %lu, "
				"len = %lu) {%s}}\n"
				, csv_str, (unsigned long)csv_buf_sz
				, (unsigned long)strlen(csv_str)
				,       ((csv_str == NULL)? "": csv_str)
				);

			if (csv_str == NULL) {
				++err_cnt;
			}
		}
	}
	printf("+++ test_string_to_csv_string(%p) finished! "
		"{err_cnt = %d}\n\n"
		,                                 (void *)t
		,           err_cnt
		);

	if (csv_str != NULL) {
		free(csv_str); csv_str	= NULL;
	}

	return;
}

static void test_strings_to_csv_line(struct test_table_t *t
									, char delimit
									)
{
	size_t	i;
	int	err_cnt = 0;
	char	*csv_line = NULL;
	size_t	csv_line_sz = (size_t)0
		, n = (sizeof(t->lines[0].strings) /
				sizeof(t->lines[0].strings[0]) - 1)
		;

	printf("+++ test_strings_to_csv_line(%p, 0x%x) starts...\n"
		,			(void *)t,  (unsigned)delimit
		);
	for (i = 0; i < (sizeof(t->lines) / sizeof(t->lines[0])); ++i) {
		printf("*** pattern %lu ***\n", (unsigned long)(i + 1));
		if ((csv_line = strings_to_csv_line(t->lines[i].strings
							, (int)n
							, csv_line
							, &csv_line_sz
							, delimit
							)) != NULL) {
			printf("\t{strings = %p, n = %lu} => "
					"CSV-line {%p (size = %lu, "
					"len = %lu)  {%s}}\n"
				,                (void *)(t->lines[i].strings)
				, (unsigned long)n, csv_line
				,	(unsigned long)csv_line_sz
				, (unsigned long)strlen(csv_line), csv_line
				);
		}
		else {
			++err_cnt;
		}
	}
	printf("+++ test_strings_to_csv_line(%p, 0x%x) finished! "
		"{err_cnt = %d}\n\n"
		,			(void *)t, (unsigned)delimit
		,	err_cnt
		);

	if (csv_line != NULL) {
		free(csv_line); csv_line	= NULL;
	}

	return;
}

static void show_Parsed_CSV_t(Parsed_CSV_t *pcp)
{
	if (pcp != NULL) {
		if (pcp->lines == NULL) {
			printf("\tpcp->lines is NULL.\n");
		}
		else {
			int	row, col;

			for (row = 0; row < pcp->line_cnt; ++row)
			for (col = 0
			; col < pcp->lines[row]->cols_cnt
			; ++col) {
				char	*value =
					get_value_parsed_csv_t(pcp
								, row
								, col);

				printf("\tpcp[%d][%d] = {%s}\n"
					,     row, col
					, ((value != NULL)?
							value: "(null)"
						)
					);
			}
		}
	}

	return;
}

int main(int ac, char **av)
{
	int		sv_ac = ac;
	char	**sv_av = av;

	ac = sv_ac; av = sv_av;

	/* show the table */
	show_table(&table); printf("\n");

	/* test for string_to_csv_string() */
	test_string_to_csv_string(&table);

	/* test for strings_to_csv_line();
	 * COMMA separated and TAB separated
	 */
	test_strings_to_csv_line(&table, DLMT_COMMA);
	test_strings_to_csv_line(&table, DLMT_TAB);

	{ /* test for string_to_csv_string(); empty cell, only CR cell
	   */
		char	*csv, *c_strs[] = {   ""	/* Empty-cell */
					, "\n"	/* Only CR cell */
					};
		size_t csv_str_sz = (size_t)0, i;

		printf("+++ test string_to_csv_string(empty and only "
				"CR cells) starts...\n"
				);
		for (i = 0
		; i < (sizeof(c_strs) / sizeof(c_strs[0]))
		; ++i) {
			if ((csv = string_to_csv_string(c_strs[i]
							, NULL
							, &csv_str_sz
							)) != NULL) {
				printf("%s cell => CSV-string { "
					"csv (csv_str_sz = %lu, "
					"len = %lu) {%s} }\n"
					, ((strlen(c_strs[i]) == 0)?
						"Empty": "Only CR")
					, (unsigned long)csv_str_sz
					, (unsigned long)strlen(csv), csv
					);
				free(csv); csv_str_sz	= (size_t)0;
			}
			else {
				printf("!!Failed in "
					"string_to_csv_string(%s cell) "
					"returns NULL.\n"
					, ((*(c_strs[i]) == '\0')?
						"Empty": "Only CR")
					);
			}
		}
		printf("+++ test string_to_csv_string(empty and only "
			"CR cells) finished!\n\n"
			);
	}

	{  /* test get_csv_line(), csv_line_to_strings();
	    * empty, only 1 cell, only CR cell, alphabets and DQ, broken
	    * zero-sized file, TSV formatted
	    */
		size_t	i;

		printf("+++ test get_csv_line() starts...\n");
		for (i = 0; i < CNT_csv_files; ++i) {
			char	*csv_line = NULL;
			size_t	csv_line_sz	= (size_t)0;
			int		is_broken = 0, lcnt = 0;
			FILE	*fp = NULL;

			if ((fp = fopen(csv_files[i], "r")) == NULL) {
				printf("***ERROR: can't open the "
					"\"%s\" CSV-file for read-only "
					"mode.\n"
					,  csv_files[i]
					);
				/* ERROR: failed in fopen(CSV-file) */
				exit(1);
			}

			printf(">>> apply get_csv_line() to the %lu-th "
				"CSV-file \"%s\" (fp = %p)...\n"
				,                               (unsigned long)i
				,           csv_files[i], (void *)fp
				);
			while ((csv_line = get_csv_line(fp
						, csv_line
						, &csv_line_sz
						, &is_broken
						)) != NULL) {
				if (is_broken != 0) {
					printf("***ERROR: the \"%s\" "
						"CSV-file is broken "
						"format "
						"(unexpected EOF, "
						"last read-in buf "
						"[L%d] = {%s}, "
						"is_broken = %d)\n"
						,           csv_files[i]
						, ++lcnt, csv_line
						, is_broken
						);

					/* this loop would break... */
					continue;
				}

				printf("[L%d] (csv_line = %p, "
					"sz = %lu, len = %lu) {%s}\n"
					, ++lcnt, csv_line
					, (unsigned long)csv_line_sz
					,               (unsigned long)strlen(csv_line)
					, csv_line
					);

				{ /* test csv_line_to_strings() */
					char	**strings = NULL;
					int	cols_cnt = 0, j;

					printf("\t>>> apply csv_line "
						"to csv_line_to_strings"
						"()...\n"
										);
					if ((strings =
						csv_line_to_strings(
							  csv_line
							, ((i + 1 ==
							CNT_csv_files)?
							  DLMT_TAB
							: DLMT_COMMA
							)
							, &cols_cnt
							)) == NULL) {
						printf("***ERROR: "
						"failed in "
						"csv_line_to_strings"
						"(%p, '%c', %p) "
						"returns NULL. "
						"{errno = %d, "
						"cols_cnt = %d}\n"
						, csv_line
						, ((i - 1 !=
							CNT_csv_files)?
							  DLMT_COMMA
							: DLMT_TAB
						)
						, (void *)&cols_cnt
						, errno,       cols_cnt
						);
						/* ERROR: failed to
						 * convert the CSV-line
						 * to vector of
						 * C-strings
						 */
						exit(3);
					}

					printf("\tCSV-line (addr = %p) => "
							"Vector of C-strings {\n"
							"\t\t(addr = %p, cols_cnt = %d)\n"
							,				csv_line
							,			(void *)strings, cols_cnt
							);
					for (j = 0; j < cols_cnt; ++j) {
						printf("\t\t\t(len = %lu) {%s}\n"
							,		(unsigned long)strlen(strings[j])
							,					strings[j]
							);
					}
					printf("\t}\n");

					free(strings);
					printf("\t>>> Done csv_line to "
											"csv_line_to_strings()...\n"
						);
				}
			}
			printf("<<< Done get_csv_line() to the %lu-th CSV-file "
					"\"%s\"! {errno = %d, fp = %p, csv_line = %p, "
					"csv_line_sz = %lu, lcnt = %d, "
					"is_broken = %d}\n"
					,                              (unsigned long)i
					,  csv_files[i],  errno, (void *)fp,	csv_line
					,			(unsigned long)csv_line_sz, lcnt
					,            is_broken
					);

			(void)fclose(fp);
		}
		printf("+++ test get_csv_line() finished!\n\n");
	}

	{  /* test parse_csv_file() and free_parsed_csv_t() */
		struct test_result_parse_csv_file_t {
			Errcode_Parse_CSV_File_t	errcode;
			Parsed_CSV_t				parsed;
			} results[CNT_csv_files];
		size_t	i;

		printf("+++ test parse_csv_file(), free_parsed_csv_t() and "
				"get_value_parsed_csv_t() starts...\n");
		memset(results, (int)'\0', sizeof(results));
		for (i = 0; i < CNT_csv_files; ++i) {
			printf(">>> Apply \"%s\" %s-file to parse_csv_file()...\n"
					,           csv_files[i], ((i + 1 != CNT_csv_files)?
												"CSV": "TSV")
					);

			results[i].parsed.line_cnt	= 0;
			results[i].parsed.lines	= NULL;
			results[i].parsed.min_cols_cnt	= 0;
			results[i].parsed.max_cols_cnt	= 0;
			results[i].errcode	=
							parse_csv_file(csv_files[i]
											, (i + 1 != CNT_csv_files)? 
												DLMT_COMMA: DLMT_TAB
											, &(results[i].parsed)
											);

			printf("\tresult [%lu]: errcode = %d\n"
					,		(unsigned long)i + 1, results[i].errcode
					);
			show_Parsed_CSV_t(&(results[i].parsed));
			printf("<<< Done \"%s\" %s-file to parse_csv_file()!\n\n"
					,          csv_files[i], ((i + 1 != CNT_csv_files)?
												"CSV": "TSV")
					);
		}

		for (i = 0; i < CNT_csv_files; ++i) {
			printf(">>> Apply the parsed structure for \"%s\""
					" %s-file to free_parsed_csv_t()...\n"
					,                                    csv_files[i]
					, ((i + 1 != CNT_csv_files)? "CSV": "TSV")
					);

			free_parsed_csv_t(&(results[i].parsed));

			printf("<<< Done the parsed structure for \"%s\""
					" %s-file to free_parsed_csv_t()!\n\n"
					,                                   csv_files[i]
					, ((i + 1 != CNT_csv_files)? "CSV": "TSV")
					);
		}

		printf("+++ test parse_csv_file(), free_parsed_csv_t() and "
				"get_value_parsed_csv_t() finished!\n");
	}

	exit(0);/* exit this program */
}
